/*
 * Copyright (C) 2001-2006 The Mir-coders group
 *
 * This file is part of Mir.
 *
 * Mir is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Mir is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Mir; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * In addition, as a special exception, The Mir-coders gives permission to link
 * the code of this program with  any library licensed under the Apache Software License,
 * and distribute linked combinations including the two.  You must obey the
 * GNU General Public License in all respects for all of the code used other than
 * the above mentioned libraries.  If you modify this file, you may extend this
 * exception to your version of the file, but you are not obligated to do so.
 * If you do not wish to do so, delete this exception statement from your version.
 */
package mir.storage;

import mir.log.LoggerWrapper;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * Base class for (sql) statement generators
 */
public class StatementGenerator {
  private static LoggerWrapper logger = new LoggerWrapper("Database.Statements");

  public StatementGenerator() {
    query = new StringBuffer();
    parameters = new ArrayList();
  }

  protected void appendQuery(String aPart) {
    query.append(aPart);
  }

  protected void appendParameter(Object aParameter) {
    parameters.add(aParameter);
  }

  /**
    * Executes the statement. Returns the number of modified records.
    */
  protected int executeWithModifiedCount(Connection aConnection) throws DatabaseFailure {
    long start = System.currentTimeMillis();

    try {
      aConnection.setAutoCommit(false);
      logQueryBefore(query.toString());

      PreparedStatement statement = aConnection.prepareStatement(query.toString());

      int index = 1;
      Iterator i = parameters.iterator();

      while (i.hasNext()) {
        statement.setObject(index, i.next());
        index++;
      }

      int result = statement.executeUpdate();

      logQueryAfter(query.toString(), System.currentTimeMillis() - start);

      return result;
    }
    catch (Throwable e) {
      logQueryError(query.toString(), System.currentTimeMillis() - start, e);
      System.out.println(e.toString());

      throw new DatabaseFailure(e);
    }
    finally {
      try {
        aConnection.setAutoCommit(true);
      }
      catch (Exception e) {
      }
    }
  }

  /**
    * Executes the statement. Returns a resultset.
    */
  protected ResultSet executeWithResultSet(Connection aConnection) throws DatabaseFailure {
    long start = System.currentTimeMillis();

    try {
      logQueryBefore(query.toString());

      PreparedStatement statement = aConnection.prepareStatement(query.toString());

      int index = 1;
      Iterator i = parameters.iterator();

      while (i.hasNext()) {
        statement.setObject(index, i.next());
        index++;
      }

      ResultSet result = statement.executeQuery();

      logQueryAfter(query.toString(), System.currentTimeMillis() - start);

      return result;
    }
    catch (Throwable e) {
      logQueryError(query.toString(), System.currentTimeMillis() - start, e);
      System.out.println(e.toString());

      throw new DatabaseFailure(e);
    }
  }

  private void logQueryBefore(String aQuery) {
    logger.debug("about to perform QUERY " + aQuery);
  }

  private void logQueryAfter(String aQuery, long aTime) {
    logger.info("QUERY " + aQuery + " took " + aTime + "ms.");
  }

  private void logQueryError(String aQuery, long aTime, Throwable anException) {
    logger.error("QUERY " + aQuery + " took " + aTime + "ms, but threw exception " + anException.toString());
  }


  private StringBuffer query;
  private List parameters;
}
